from typing import Any, List, Callable, Dict, Optional, AsyncGenerator, Tuple, Union, Type
from motor.motor_asyncio import AsyncIOMotorCollection, AsyncIOMotorDatabase, AsyncIOMotorCursor

from motorized.query import Q
from motorized.client import connection
from motorized.exceptions import NotConnectedException


class QuerySet:
    def __init__(self, model):
        self.model = model
        self._query = Q()
        self._limit = None
        self._sort = None
        self.database = None

    def copy(self) -> "QuerySet":
        instance = QuerySet(self.model)
        instance._query = self._query.copy()
        instance._limit = self._limit
        instance._sort = self._sort
        instance.use(self.database)
        return instance

    def __repr__(self):
        return f'<QuerySet: {self.model.__name__}: {self._query}>'

    @classmethod
    def from_query(cls, model: Type["Document"], query: Q) -> "QuerySet":
        instance = QuerySet(model)
        instance._query = query
        return instance

    def use(self, database: Optional[AsyncIOMotorDatabase]) -> None:
        """if `database` is None then the default database will be used.
        """
        self.database = database

    async def __aiter__(self) -> AsyncGenerator["Document", None]:
        cursor = await self.find()
        async for data in cursor:
            instance = self.model(**data)
            yield instance

    async def create(self, *args, **kwargs) -> "Document":
        instance = self.model(*args, **kwargs)
        await instance.save()
        return instance

    @property
    def collection(self) -> AsyncIOMotorCollection:
        if self.database is not None:
            return getattr(self.database, self.model.Mongo.collection)
        if not connection.database:
            raise NotConnectedException('You need to use connection.connect before using collection')
        return getattr(connection.database, self.model.Mongo.collection)

    async def count(self, **kwargs) -> int:
        return await self.collection.count_documents(self._query.query, **kwargs)

    async def first(self) -> Optional["Document"]:
        try:
            return await self.__aiter__().__anext__()
        except StopAsyncIteration:
            return None

    async def drop(self) -> None:
        await self.collection.drop()

    def filter(self, **kwargs) -> "QuerySet":
        instance = self.copy()
        instance._query += Q(**kwargs)
        return instance

    def exclude(self, **kwargs) -> "QuerySet":
        instance = self.copy()
        query = Q()
        query.query = Q.convert_kwargs_to_query(**kwargs, invert=True)
        instance._query += query
        return instance

    def limit(self, size: int) -> "QuerySet":
        instance = self.copy()
        instance._limit = size
        return instance

    def sort(self, ordering: Optional[Union[str, List[str]]]) -> "QuerySet":
        instance = self.copy()
        if ordering:
            instance._sort = self._sort_instruction(ordering if hasattr(ordering, '__iter__') else [ordering])
        else:
            instance._sort = None
        return instance

    def __add__(self, other: "QuerySet") -> "QuerySet":
        instance = self.copy()
        instance._query += other._query
        return instance

    async def all_list(self) -> List["Document"]:
        return [instance async for instance in self]

    async def map(self, func: Callable) -> None:
        async for instance in self.all():
            await func(instance)

    async def distinct(self, key: str, **kwargs) -> List[Any]:
        if not self._query.is_empty():
            kwargs.setdefault('filter', self._query.query)
        return await self.collection.distinct(key, **kwargs)

    async def aggregate(self, pipeline, **kwargs) -> Any:
        return await self.collection.aggregate(pipeline, **kwargs)

    async def find(self, **kwargs) -> AsyncIOMotorCursor:
        cursor = self.collection.find(self._query.query, **kwargs)
        return await self._paginate_cursor(cursor)

    async def find_one(self) -> Dict:
        return await self.collection.find_one(self._query.query)

    async def delete_one(self):
        return await self.collection.delete_one(filter=self._query.query)

    async def get(self, **kwargs) -> "Document":
        instance = self.filter(**kwargs)
        cursor = await instance.find()
        try:
            first = await cursor.__anext__()
        except StopAsyncIteration as error:
            raise self.model.DocumentNotFound(self._query.query) from error
        try:
            second = await cursor.__anext__()
        except StopAsyncIteration:
            return self.model(**first)

        raise self.model.TooManyMatchException

    def fresh(self) -> "QuerySet":
        """Return a fresh queryset without any filtering/ordering/limiting parameter
        as fresh as new.
        """
        instance = QuerySet(self.model)
        instance.use(self.database)
        return instance

    async def values_list(self, fields: List[str], flat=False, noid=False):
        if isinstance(fields, str):
            fields = (fields,)
        projection = {f: True for f in fields}
        if noid:
            projection['_id'] = False

        cursor: AsyncIOMotorCursor = self.collection.find(projection=projection)
        if not flat:
            return list([item async for item in cursor])
        assert len(fields) == 1, 'You can only have one field using flat=True'
        field_name = fields[0]
        return list([value[field_name] async for value in cursor])

    async def _paginate_cursor(self, cursor: AsyncIOMotorCursor) -> AsyncIOMotorCursor:
        if self._sort:
            cursor = cursor.sort(self._sort)
        if self._limit:
            cursor = cursor.limit(self._limit)
        return cursor

    @staticmethod
    def _sort_instruction(order: List[str]) -> List[Tuple[str, int]]:
        """Convert a list for format:
        ['name', '-age'] to:
        [('name': 1), ('age': -1)]
        """
        def generate_tuple(word) -> Tuple[str, int]:
            if word.startswith('-'):
                return (word[1:], -1)
            return (word, 1)

        return [generate_tuple(word) for word in order]

    async def exists(self) -> bool:
        return await self.collection.count_documents(self._query.query, limit=1) > 0

    async def delete(self, **kwargs):
        return await self.collection.delete_many(self._query.query, **kwargs)

    async def delete_one(self, **kwargs):
        return await self.collection.delete_one(self._query.query, **kwargs)

    async def pop(self, **kwargs) -> "Document":
        instance = self.filter(**kwargs)
        document_dict = await instance.collection.find_one_and_delete(instance.query._query)
        document = self.model(**document_dict)
        return document

    async def indexes(self):
        return [index async for index in self.collection.list_indexes()]



from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorCollection, AsyncIOMotorDatabase
from typing import Optional, Union, Any, Optional, Dict, Type, List, Generator
from pydantic import BaseModel, Field
from pydantic.fields import ModelField
from pydantic.main import ModelMetaclass
from pymongo.results import InsertOneResult, UpdateResult

from motorized.queryset import QuerySet
from motorized.query import Q
from motorized.types import PydanticObjectId, ObjectId
from motorized.exceptions import DocumentNotSavedError, MotorizedError


def show_class_constructor(cls, name, bases, optdict: Dict):
    print(' --- ')
    print('cls', cls)
    print('name', name)
    print('bases:', *bases)
    print('opts', optdict)


class DocumentMeta(ModelMetaclass):
    def __new__(cls, name, bases, optdict: Dict) -> Type['Document']:
        # optdict.pop('objects', None)
        # optdict.pop('__annotations__', {}).pop('objects', None)
        # show_class_constructor(cls, name, bases, optdict)
        instance: Type[Document] = super().__new__(cls, name, bases, optdict)
        if name not in ('Document',):
            cls._populate_default_mongo_options(cls, name, instance, optdict.get('Mongo'))

        class DocumentError(MotorizedError):
            pass

        class TooManyMatchException(DocumentError):
            pass

        class DocumentNotFound(DocumentError):
            pass

        instance.DocumentError = DocumentError
        instance.TooManyMatchException = TooManyMatchException
        instance.DocumentNotFound = DocumentNotFound
        instance.objects = instance.Mongo.manager_class(instance)
        return instance

    def _populate_default_mongo_options(cls, name: str, instance: "Document",
                                        custom_mongo_settings_class) -> None:
        class Mongo:
            pass

        # forbid re-utilisation of the Mongo class between inheritance of the class
        try:
            if instance.Mongo.class_name != name:
                instance.Mongo = Mongo()
        except AttributeError:
            pass

        if custom_mongo_settings_class:
            instance.Mongo = custom_mongo_settings_class

        default_settings = {
            'collection': name.lower() + 's',
            'manager_class': QuerySet,
            'class_name': name
        }

        for attribute_name, default_value in default_settings.items():
            if not hasattr(instance.Mongo, attribute_name):
                setattr(instance.Mongo, attribute_name, default_value)


class Document(BaseModel, metaclass=DocumentMeta):
    # objects: QuerySet
    id: Optional[PydanticObjectId] = Field(alias='_id')

    class Config:
        json_encoders = {ObjectId: str}

    class Mongo:
        manager_class = QuerySet

    def __init__(self, *args, **kwargs) -> None:
        BaseModel.__init__(self, *args, **self._transform(**kwargs))

    def get_query(self) -> Q:
        document_id = getattr(self, 'id', None)
        if not document_id:
            raise DocumentNotSavedError('document has no id.')
        return Q(_id=document_id)

    async def _create(self, creation_dict: Dict) -> InsertOneResult:
        response = await self.objects.collection.insert_one(creation_dict)
        self.id = response.inserted_id
        return response

    async def _update(self, update_dict: Dict) -> UpdateResult:
        return await self.objects.collection.update_one(
            filter={'_id': self.id},
            update={'$set': update_dict}
        )

    async def to_mongo(self) -> Dict:
        """Convert the current model dictionary to database output dict,
        this also mean the aliased fields will be stored in the alias name instead of their
        name in the document declaration.
        """
        saving_data = self.dict()
        for field in self._aliased_fields():
            saving_data[field.alias] = saving_data.pop(field.name, None)
        return saving_data

    async def save(self) -> Union[InsertOneResult, UpdateResult]:
        data = await self.to_mongo()
        document_id = data.pop('_id', None)
        if document_id is None:
            return await self._create(data)
        return await self._update(data)

    async def commit(self) -> "Document":
        """Same as `.save` but return the current instance.
        """
        await self.save()
        return self

    async def delete(self) -> "Document":
        """Delete the current instance from the database,
        to the deleted the instance need to have a .id set, in any case the function
        will return the instance itself
        """
        try:
            qs = self.objects.from_query(self, self.get_query())
            await qs.delete_one()
        except DocumentNotSavedError:
            pass
        setattr(self, 'id', None)
        return self

    async def fetch(self) -> "Document":
        """Return a fresh instance of the current document from the database.
        """
        qs = self.objects.copy()
        qs._query = self.get_query()
        return await qs.get()

    @classmethod
    def _aliased_fields(cls) -> Generator[List[ModelField], None, None]:
        """Return the list of fields with aliases
        """
        return [field for field in cls.__fields__.values() if field.name != field.alias]

    def _transform(self, **kwargs) -> Dict:
        """Override this method to change the input database before having it
        being validated/parsed by BaseModel (pydantic)
        """
        return kwargs



from bson import ObjectId
from pydantic.main import BaseModel


class PydanticObjectId(ObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, ObjectId):
            raise TypeError('ObjectId required')
        return v

    @classmethod
    def __modify_schema__(cls, schema: dict):
        schema['type'] = 'string'



from typing import List, Any, Dict
from motorized.keywords import Eq, Neq, In, Nin, Gte, Lte, Gt, Lt, Exists, Regex
from motorized.utils import merge_values, dict_deep_update


KEYWORDS = {
    'eq': Eq,
    'neq': Neq,
    'in': In,
    'nin': Nin,
    'gte': Gte,
    'lte': Lte,
    'gt': Gt,
    'lt': Lt,
    'exists': Exists,
    'regex': Regex
}


class Q:
    def __init__(self, **kwargs) -> None:
        self.query = self.convert_kwargs_to_query(**kwargs)

    def __repr__(self):
        return f'<Q: {self.query}>'

    def copy(self):
        instance = Q()
        instance.query = self.query.copy()
        return instance

    @classmethod
    def raw(self, query: Dict) -> "Q":
        instance = Q()
        instance.query = query
        return instance

    @classmethod
    def convert_kwargs_to_query(cls, invert=False, **kwargs) -> Dict:
        query = {}
        for key, value in kwargs.items():
            path, value = cls.apply_keywords(value, key.split('__'),
                                              invert=invert)
            filter_dict = cls.dict_path(path, value)
            dict_deep_update(query, filter_dict,
                             on_conflict=merge_values)
        return query

    @staticmethod
    def read_dict_path(data: dict, path: List['str']) -> Any:
        x = data
        for node in path:
            x = x[node]
        return x

    @staticmethod
    def dict_path(path: List[str], value: Any = None) -> Dict:
        """Construct a dictionary from a path to hold the given value,
        example:
        d = QuerySet.dict_path(['a', 'b', 'c'], 42)
        d == {'a': {'b': {'c': 42}}}
        """
        out = {}
        node = out
        last_node = None
        last_key = None
        for k in path:
            node[k] = {}
            last_node = node
            node = node[k]
            last_key = k
        last_node[last_key] = value
        return out

    @classmethod
    def apply_keywords(cls, raw_value, path: List[str], invert=False):
        if invert and path[-1] not in KEYWORDS:
            path.append('eq')

        for cmd in path:
            try:
                op = KEYWORDS[cmd](raw_value)
                return path[0:-1], op.as_mongo_expression(invert)
            except KeyError:
                pass
        return path, raw_value

    def __add__(self, other: "Q") -> "Q":
        instance = self.copy()
        dict_deep_update(instance.query, other.query, on_conflict=merge_values)
        return instance

    def is_empty(self) -> bool:
        return not self.query

    def __eq__(self, other: "Q") -> bool:
        return self.query == other.query
